/*
 * JBoss, the OpenSource J2EE webOS
 *
 * Distributable under LGPL license.
 * See terms of license at gnu.org.
 */
package org.jboss.mail.store;

import java.io.IOException;
import java.io.InputStream;
import java.nio.ByteBuffer;
import java.sql.SQLException;

import javax.naming.NamingException;


/**
 * A stream that reads data from a Blob in a disconnected manner.    Responible for managing
 * which store (id) and where in the store (position) the data is to come from.
 * 
 * @author Michael Barker <mailto:mike@middlesoft.co.uk>
 *
 */
public class BlobInputStream extends InputStream
{
   protected final static int DEFAULT_BUFFER_SIZE = 1024;
   
   private ByteBuffer buffer;
   private int minBufferSize;
   private int maxBufferSize;
   private int bufferSize;
   private long position;
   private Store store;

   private StoreItemMetaData meta;
   
   /**
    * Initialises the input stream.
    * 
    * @param dsName
    * @param id
    * @param bufferSize
    * @param create
    * @throws StoreException 
    * @throws StoreItemNotFoundException
    * @throws NamingException
    * @throws SQLException
    */
   public BlobInputStream(Store store, StoreItemMetaData meta, int minBufferSize, int maxBufferSize) throws StoreException
   {
      this.buffer = null;
      this.store = store;
      this.meta = meta;
      this.bufferSize = Math.max(Math.min(maxBufferSize, longToInt(meta.getSize())), minBufferSize); 
      this.maxBufferSize = maxBufferSize;
      this.minBufferSize = minBufferSize;
      this.position = meta.getStartIndex();
   }   
  
   
   private static int longToInt(long l)
   {
      if (l >= Integer.MAX_VALUE)
      {
         return Integer.MAX_VALUE;
      }
      else if (l <= Integer.MIN_VALUE)
      {
         return Integer.MIN_VALUE;
      }
      else
      {
         return (int) l;
      }
   }
   


   /**
    * Read a byte at a time from the buffer for the blob.
    * 
    * @see java.io.InputStream#read()
    */
   public int read() throws IOException
   {
      boolean isRead = false;
      
      // If the buffer has not been initialised.
      // If all of the bytes have been read from the buffer.
      if (buffer == null || !buffer.hasRemaining())
      {
         int i = loadBuffer();
         isRead = (i == 0);
      }
      
      // If the entire blob has be read exit.
      if (isRead)
      {
         return -1;
      }
      else
      {
         return buffer.get() & 0x000000ff;         
      }
   }
   
   
   public int read(byte[] b, int off, int len) throws IOException
   {
      int fromBuffer = 0;
      int fromStore = 0;
      
      if (buffer == null || !buffer.hasRemaining())
      {
         loadBuffer();
      }
      
      // Copy what is currently in the buffer to the output.
      if (buffer.hasRemaining())
      {
         fromBuffer = Math.min(buffer.remaining(), len);
         buffer.get(b, off, fromBuffer);
         // If there is still more bytes to be read, read direct
         // from the store.
         if (fromBuffer < len)
         {
            fromStore = doRead(b, off + fromBuffer, len - fromBuffer);
         }
         return fromBuffer + fromStore;
      }
      else
      {
         return -1;
      }      
   }
   
   
   private int doRead(byte[] b, int off, int len) throws IOException
   {
      try
      {
         int numBytes = store.read(b, off, len, meta.getId(), position);
         if (numBytes > 0)
         {
            position += numBytes;            
         }
         return numBytes;
      }
      catch (StoreException e)
      {
         throw new IOException(e.getMessage());
      }
      
   }

   /**
    * Loads and wraps bytes directly from the store.
    * Does less copying.
    * 
    * @return
    * @throws IOException
    */
   private int loadBuffer() throws IOException
   {
      
      try
      {
         //byte[] b = store.getBytes(meta.getId(), position, maxBufferSize);
         //this.buffer = ByteBuffer.wrap(b);
         this.buffer = store.getBuffer(meta.getId(), position, maxBufferSize);
         position += buffer.limit();
         return buffer.limit();
      }
      catch (StoreException e)
      {
         throw new IOException(e.getMessage());
      }
   }

   
}

