/*
 * JBoss, the OpenSource J2EE webOS
 *
 * Distributable under LGPL license.
 * See terms of license at gnu.org.
 */
package org.jboss.mail.store;

import java.io.IOException;
import java.io.OutputStream;
import java.nio.ByteBuffer;
import java.sql.SQLException;

import javax.naming.NamingException;

/**
 * Input stream for write to underlying data stores.  Responible for managing
 * which store (id) and where in the store (position) the data is to go.
 * 
 * @author Michael Barker <mailto:mike@middlesoft.co.uk>
 *
 */
public class BlobOutputStream extends OutputStream
{
   //protected final static String CREATE_STREAM = "INSERT INTO STREAM_STORE VALUES (ID, DATA) VALUES (?, ?)";
   protected final static int DEFAULT_BUFFER_SIZE = 1024;
   
   private int bufferSize;
   private ByteBuffer buffer;
   private Store store;
   private long position;

   private StoreItemMetaData meta;
   
   
   /**
    * Initialises an input stream and will load the first <code>bufferSize</code>
    * bytes into the buffer.  If the <code>create</code> is true, it will try
    * to create the buffer before hand.
    * 
    * @param dsName
    * @param id
    * @param bufferSize
    * @param create
    * @throws StoreItemNotFoundException
    * @throws NamingException
    * @throws SQLException
    */
   public BlobOutputStream(Store store, StoreItemMetaData meta, int bufferSize)
   {
      this.store = store;
      this.meta = meta;
      this.buffer = ByteBuffer.allocate(bufferSize);
      this.position = meta.getStartIndex();
   }


   

   /* (non-Javadoc)
    * @see java.io.OutputStream#write(int)
    */
   public void write(int b) throws IOException
   {
      // Flush the buffer if it is full.
      if (buffer.remaining() <= 0)
      {
         flush();
      }
      buffer.put((byte) b);
   }
   
   /**
    * Implement batch writing to the store.  Basically flush
    * whatever is in the current buffer and then write this
    * entire array to the store.
    */
   public void write(byte[] b, int off, int len) throws IOException
   {
      // flush current buffer.
      flush();
      // Store the entire array now.
      doWrite(b, off, len);      
   }
   

   
   private int doWrite(byte[] b, int off, int len) throws IOException
   {
      try
      {
         int numBytes = store.write(b, off, len, meta.getId(), position);
         position += numBytes;
         return numBytes;
      }
      catch (Exception e)
      {
         throw new IOException(e.getMessage());
      }
      
   }
   
   
   public void flush() throws IOException
   {
      if (buffer.position() > 0)
      {
         buffer.flip();
         doWrite(buffer.array(), buffer.position(), buffer.remaining());
         buffer.clear();
      }
   }
   
   
   public void close() throws IOException
   {
      flush();
   }


}
