/*
 * JBoss, the OpenSource J2EE webOS
 *
 * Distributable under LGPL license.
 * See terms of license at gnu.org.
 */
package org.jboss.mail.store.bdb;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.nio.ByteBuffer;
import java.util.zip.GZIPInputStream;
import java.util.zip.GZIPOutputStream;

import org.jboss.mail.store.AbstractStore;
import org.jboss.mail.store.BlobInputStream;
import org.jboss.mail.store.BlobOutputStream;
import org.jboss.mail.store.Store;
import org.jboss.mail.store.StoreException;
import org.jboss.mail.store.StoreItem;
import org.jboss.mail.store.StoreItemImpl;
import org.jboss.mail.store.StoreItemMetaData;
import org.jboss.mail.store.StoreItemNotFoundException;

import com.sleepycat.je.Database;
import com.sleepycat.je.DatabaseConfig;
import com.sleepycat.je.DatabaseEntry;
import com.sleepycat.je.DatabaseException;
import com.sleepycat.je.Environment;
import com.sleepycat.je.EnvironmentConfig;
import com.sleepycat.je.LockMode;
import com.sleepycat.je.OperationStatus;

/**
 * @author Michael Barker <mailto:mike@middlesoft.co.uk>
 *
 */
public class BDBStore extends AbstractStore implements Store
{
   private Database db;
   private String home;

   public BDBStore()
   {
   }
   
   public void setHome(String home)
   {
      this.home = home;
   }
   
   public String getHome()
   {
      return home;
   }
   
   /* (non-Javadoc)
    * @see org.jboss.mail.store.Store#read(byte[], int, int, java.lang.Object, long)
    */
   public ByteBuffer getBuffer(Object id, long position, int len) throws StoreException
   {
      DatabaseEntry key = new DatabaseEntry(BDBStoreUtil.toBytes(id));
      DatabaseEntry data = new DatabaseEntry();
      ByteBuffer buffer;
      try
      {
         data.setPartial(true);
         data.setPartialOffset((int) position);
         data.setPartialLength(len);
         
         OperationStatus opStatus = db.get(null, key, data, LockMode.DEFAULT);
         if (OperationStatus.SUCCESS.equals(opStatus))
         {   
            buffer = ByteBuffer.wrap(data.getData());
         }
         else if (OperationStatus.NOTFOUND.equals(opStatus))
         {
            throw new StoreItemNotFoundException(id);
         }
         else
         {
            throw new StoreException("Database invalid operation status: " + opStatus.toString());
         }
      }
      catch (DatabaseException e)
      {
         throw new StoreException(e);
      }
      
      return buffer;
   }

   /* (non-Javadoc)
    * @see org.jboss.mail.store.Store#read(byte[], int, int, java.lang.Object, long)
    */
   public long getSize(Object id) throws StoreException
   {
      DatabaseEntry key = new DatabaseEntry(BDBStoreUtil.toBytes(id));
      DatabaseEntry data = new DatabaseEntry();
      try
      {
         OperationStatus opStatus = db.get(null, key, data, LockMode.DEFAULT);
         if (OperationStatus.SUCCESS.equals(opStatus))
         {
            return data.getSize();
         }
         else if (OperationStatus.NOTFOUND.equals(opStatus))
         {
            throw new StoreItemNotFoundException(id);
         }
         else
         {
            throw new StoreException("Database invalid operation status: " + opStatus.toString());
         }
      }
      catch (DatabaseException e)
      {
         throw new StoreException(e);
      }
   }

   

   /* (non-Javadoc)
    * @see org.jboss.mail.store.Store#read(byte[], int, int, java.lang.Object, long)
    */
   public int read(byte[] b, int off, int len, Object id, long position) throws StoreException
   {
      DatabaseEntry key = new DatabaseEntry(BDBStoreUtil.toBytes(id));
      DatabaseEntry data = new DatabaseEntry();
      try
      {
         data.setPartial(true);
         data.setPartialOffset((int) position);
         data.setPartialLength(len);
         OperationStatus opStatus = db.get(null, key, data, LockMode.DEFAULT);
         if (OperationStatus.SUCCESS.equals(opStatus))
         {
            byte[] dataArr = data.getData();
            System.arraycopy(dataArr, 0, b, off, len);
            return dataArr.length;
         }
         else if (OperationStatus.NOTFOUND.equals(opStatus))
         {
            throw new StoreItemNotFoundException(id);
         }
         else
         {
            throw new StoreException("Database invalid operation status: " + opStatus.toString());
         }
      }
      catch (DatabaseException e)
      {
         throw new StoreException(e);
      }
   }

   /* (non-Javadoc)
    * @see org.jboss.mail.store.Store#write(byte[], int, int, java.lang.Object, long)
    */
   public int write(byte[] b, int off, int len, Object id, long position) throws StoreException
   {
      DatabaseEntry key = new DatabaseEntry(BDBStoreUtil.toBytes(id));
      DatabaseEntry data = new DatabaseEntry();
      try
      {
         data.setPartial(true);
         data.setPartialOffset((int) position);
         data.setPartialLength(len);
         data.setData(b, off, len);
         
         OperationStatus opStatus = db.put(null, key, data);
         
         if (OperationStatus.SUCCESS.equals(opStatus))
         {
            return len;
         }
         else if (OperationStatus.NOTFOUND.equals(opStatus))
         {
            throw new StoreItemNotFoundException(id);
         }
         else
         {
            throw new StoreException("Database invalid operation status: " + opStatus.toString());
         }
      }
      catch (DatabaseException e)
      {
         throw new StoreException(e);
      }
   }

   /* (non-Javadoc)
    * @see org.jboss.mail.store.Store#read(java.nio.ByteBuffer, java.lang.Object, long)
    */
   public int read(ByteBuffer buffer, Object id, long position) throws StoreException
   {
      DatabaseEntry key = new DatabaseEntry(BDBStoreUtil.toBytes(id));
      DatabaseEntry data = new DatabaseEntry();
      try
      {
         data.setPartial(true);
         data.setPartialOffset((int) position);
         data.setPartialLength(buffer.remaining());
         
         OperationStatus opStatus = db.get(null, key, data, LockMode.DEFAULT);
         if (OperationStatus.SUCCESS.equals(opStatus))
         {
            byte[] dataArr = data.getData();
            buffer.put(dataArr);
            return dataArr.length;
         }
         else if (OperationStatus.NOTFOUND.equals(opStatus))
         {
            throw new StoreItemNotFoundException(id);
         }
         else
         {
            throw new StoreException("Database invalid operation status: " + opStatus.toString());
         }
      }
      catch (DatabaseException e)
      {
         throw new StoreException(e);
      }
   }

   /* (non-Javadoc)
    * @see org.jboss.mail.store.Store#write(java.nio.ByteBuffer, java.lang.Object, long)
    */
   public int write(ByteBuffer buffer, Object id, long position) throws StoreException
   {
      if (buffer.hasArray())
      {
         int numBytes = write(buffer.array(), buffer.position(), buffer.remaining(), id, position);
         buffer.position(buffer.position() + numBytes);
         return numBytes;
      }
      
      DatabaseEntry key = new DatabaseEntry(BDBStoreUtil.toBytes(id));
      DatabaseEntry data = new DatabaseEntry();
      try
      {
         data.setPartial(true);
         data.setPartialOffset((int) position);
         int len = buffer.remaining();
         data.setPartialLength(len);
         byte[] byteArr = new byte[len];
         buffer.get(byteArr);
         data.setData(byteArr);
         
         OperationStatus opStatus = db.put(null, key, data);
         
         if (OperationStatus.SUCCESS.equals(opStatus))
         {
            return len;
         }
         else if (OperationStatus.NOTFOUND.equals(opStatus))
         {
            throw new StoreItemNotFoundException(id);
         }
         else
         {
            throw new StoreException("Database invalid operation status: " + opStatus.toString());
         }
      }
      catch (DatabaseException e)
      {
         throw new StoreException(e);
      }
   }


   /**
    * Berkley DB is 0 based.
    * 
    * @see org.jboss.mail.store.Store#getStartIndex()
    */
   public long getStartIndex()
   {
      return 0;
   }


   /* (non-Javadoc)
    * @see org.jboss.mail.store.AbstractStore#init()
    */
   public void init() throws StoreException
   {
      try
      {
         File dir = new File(home);
         if (!dir.exists())
         {
            dir.mkdirs();
         }
         
         EnvironmentConfig config = new EnvironmentConfig();
         config.setAllowCreate(true);
         Environment env = new Environment(dir, config);
         
         DatabaseConfig dbConfig = new DatabaseConfig();
         dbConfig.setAllowCreate(true);
         
         this.db = env.openDatabase(null, "jboss-mail.store", dbConfig);         
      }
      catch (Exception e)
      {
         throw new StoreException("Unable to open database at: " + home);
      }
   }

   public void validate() throws StoreException
   {
      super.validate();
      
      File f = new File(home);
      if (f.exists() && !f.isDirectory())
      {
         throw new StoreException("The home directory " + getHome() + " is not a directory");
      }      
   }
   

   /* (non-Javadoc)
    * @see org.jboss.mail.store.Store#createStoreItem(java.lang.Object)
    */
   public StoreItem createStoreItem() throws StoreException
   {
      Object id = createId();
      byte compressFlag = BDBStoreUtil.booleanToByte(isCompress());
      DatabaseEntry key = new DatabaseEntry(BDBStoreUtil.toBytes(id));
      DatabaseEntry data = new DatabaseEntry(new byte[] { compressFlag });
      StoreItem item;
      try
      {
         OperationStatus opStatus = db.put(null, key, data);
         if (!OperationStatus.SUCCESS.equals(opStatus))
         {
            throw new StoreException("Unable to create stream operation status: " + opStatus);
         }
         item = new StoreItemImpl(id, this);
      }
      catch (DatabaseException e)
      {
         throw new StoreException(e);
      }
      
      return item;
   }


   /* (non-Javadoc)
    * @see org.jboss.mail.store.Store#getInputStream(java.lang.Object)
    */
   public InputStream getInputStream(Object id) throws StoreException
   {
      InputStream in;
      
      StoreItemMetaData meta = getStoreItemMetaData(id);

      BlobInputStream blobIn = new BlobInputStream(this, meta, getMinBufferSize(), getMaxBufferSize());
      if (meta.isCompressed())
      {
         try
         {
            in = new GZIPInputStream(blobIn, getCompressBufferSize());
         }
         catch (IOException e)
         {
            throw new StoreException("Unable to create compressed stream for id =  " + id, e);
         }
      }
      else
      {
         in = blobIn;
      }
      
      return in;
   }


   /* (non-Javadoc)
    * @see org.jboss.mail.store.Store#getOuputStream(java.lang.Object)
    */
   public OutputStream getOuputStream(Object id) throws StoreException
   {
      OutputStream out;
      StoreItemMetaData meta = getStoreItemMetaData(id);
      
      BlobOutputStream blobOut = new BlobOutputStream(this, meta, getBufferSize());
      if (meta.isCompressed())
      {
         try
         {
            out = new GZIPOutputStream(blobOut, getCompressBufferSize());
         }
         catch (IOException e)
         {
            throw new StoreException("Unable to create compressed stream for id = " + id, e);
         }
      }
      else
      {
         out = blobOut;
      }
      
      return out;
   }


   /* (non-Javadoc)
    * @see org.jboss.mail.store.Store#getStoreItemMetaData(java.lang.Object)
    */
   public StoreItemMetaData getStoreItemMetaData(Object id) throws StoreException
   {
      StoreItemMetaData meta;
      DatabaseEntry key = new DatabaseEntry(BDBStoreUtil.toBytes(id));
      DatabaseEntry data = new DatabaseEntry();
      try
      {
         data.setPartial(true);
         data.setPartialOffset(0);
         data.setPartialLength(1);
         
         OperationStatus opStatus = db.get(null, key, data, LockMode.DEFAULT);
         if (OperationStatus.SUCCESS.equals(opStatus))
         {
            int size = data.getSize();
            byte[] dataArr = data.getData();
            meta = new StoreItemMetaData(id, size, BDBStoreUtil.byteToBoolean(dataArr[0]), 1L);
         }
         else if (OperationStatus.NOTFOUND.equals(opStatus))
         {
            throw new StoreItemNotFoundException(id);
         }
         else
         {
            throw new StoreException("Database invalid operation status: " + opStatus.toString());
         }
      }
      catch (DatabaseException e)
      {
         throw new StoreException(e);
      }
      return meta;
   }

}
