/*
 * JBoss, the OpenSource J2EE webOS
 *
 * Distributable under LGPL license.
 * See terms of license at gnu.org.
 */
package org.jboss.mail.store.jdbc3;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.nio.ByteBuffer;
import java.sql.Blob;
import java.sql.Connection;
import java.sql.DatabaseMetaData;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.zip.GZIPInputStream;
import java.util.zip.GZIPOutputStream;

import javax.sql.DataSource;

import org.apache.commons.dbutils.DbUtils;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.jboss.mail.store.AbstractStore;
import org.jboss.mail.store.BlobInputStream;
import org.jboss.mail.store.BlobOutputStream;
import org.jboss.mail.store.StoreClerk;
import org.jboss.mail.store.StoreException;
import org.jboss.mail.store.StoreItem;
import org.jboss.mail.store.StoreItemImpl;
import org.jboss.mail.store.StoreItemMetaData;
import org.jboss.mail.store.StoreItemNotFoundException;

/**
 * A JDBC3 Implemenation of a store that provides a backing
 * for the stream store component.
 * 
 * @author Michael Barker <mailto:mike@middlesoft.co.uk>
 *
 */
public class JDBC3Store extends AbstractStore
{
   public static final int DEFAULT_BUFFER_SIZE = 8128;
   public static final int DEFAULT_BLOCK_SIZE = 1024 * 1024;
   public static final String TRUE = "T";
   public static final String FALSE = "F";
   private static Log log = LogFactory.getLog(JDBC3Store.class);
   
   private DataSource dataSource = null;
   private int maxBlockSize = DEFAULT_BLOCK_SIZE;
   private String readStatement = null;
   private String writeStatement = null;
   private String insertStatement = null;
   private String createStatement = null;
   private String idColumn = null;
   private String blobColumn = null;
   private String compressedColumn = null;
   private String tableName = null;
   private boolean useStreams = true;
   private boolean disconnected = false;


   
   public static JDBC3Store create(DataSource ds, int maxBlockSize, String insert, String select, String blobColumn)
   {
      JDBC3Store store = new JDBC3Store();
      store.setDataSource(ds);
      //this.maxBlockSize = maxBlockSize;
      //this.select = select;
      //this.blobColumn = blobColumn;
      //this.insert = insert;
      return store;
   }
   
   
   //=====================
   // Properties
   //=====================   
   
   /**
    * @return Returns the blobColumn.
    */
   public String getBlobColumn()
   {
      return blobColumn;
   }
   /**
    * @param blobColumn The blobColumn to set.
    */
   public void setBlobColumn(String blobColumn)
   {
      this.blobColumn = blobColumn;
   }
   /**
    * @return Returns the compressedColumn.
    */
   public String getCompressedColumn()
   {
      return compressedColumn;
   }
   /**
    * @param compressedColumn The compressedColumn to set.
    */
   public void setCompressedColumn(String compressedColumn)
   {
      this.compressedColumn = compressedColumn;
   }
   /**
    * @return Returns the createStatement.
    */
   public String getCreateStatement()
   {
      return createStatement;
   }
   /**
    * @param createStatement The createStatement to set.
    */
   public void setCreateStatement(String createStatement)
   {
      this.createStatement = createStatement;
   }
   /**
    * @return Returns the dataSource.
    */
   public DataSource getDataSource()
   {
      return dataSource;
   }
   /**
    * @param dataSource The dataSource to set.
    */
   public void setDataSource(DataSource dataSource)
   {
      this.dataSource = dataSource;
   }
   /**
    * @return Returns the idColumn.
    */
   public String getIdColumn()
   {
      return idColumn;
   }
   /**
    * @param idColumn The idColumn to set.
    */
   public void setIdColumn(String idColumn)
   {
      this.idColumn = idColumn;
   }
   /**
    * @return Returns the insertStatement.
    */
   public String getInsertStatement()
   {
      return insertStatement;
   }
   /**
    * @param insertStatement The insertStatement to set.
    */
   public void setInsertStatement(String insertStatement)
   {
      this.insertStatement = insertStatement;
   }
   /**
    * @return Returns the maxBlockSize.
    */
   public int getMaxBlockSize()
   {
      return maxBlockSize;
   }
   /**
    * @param maxBlockSize The maxBlockSize to set.
    */
   public void setMaxBlockSize(int maxBlockSize)
   {
      this.maxBlockSize = maxBlockSize;
   }
   /**
    * @return Returns the readStatement.
    */
   public String getReadStatement()
   {
      return readStatement;
   }
   /**
    * @param readStatement The readStatement to set.
    */
   public void setReadStatement(String selectStatement)
   {
      this.readStatement = selectStatement;
   }
   /**
    * @return Returns the tableName.
    */
   public String getTableName()
   {
      return tableName;
   }
   /**
    * @param tableName The tableName to set.
    */
   public void setTableName(String tableName)
   {
      this.tableName = tableName;
   }
   
   /**
    * @return Returns the useStreams.
    */
   public boolean isUseStreams()
   {
      return useStreams;
   }
   /**
    * @param useStreams The useStreams to set.
    */
   public void setUseStreams(boolean useStreams)
   {
      this.useStreams = useStreams;
   }
      
   /**
    * @return Returns the disconnected.
    */
   public boolean isDisconnected()
   {
      return disconnected;
   }
   /**
    * @param disconnected The disconnected to set.
    */
   public void setDisconnected(boolean disconnected)
   {
      this.disconnected = disconnected;
   }

   /**
    * @return Returns the writeStatement.
    */
   public String getWriteStatement()
   {
      return writeStatement;
   }
   /**
    * @param writeStatement The writeStatement to set.
    */
   public void setWriteStatement(String writeStatement)
   {
      this.writeStatement = writeStatement;
   }
   
   /**
    * Get the size of the blob.
    */
   public long getSize(Object id) throws StoreException
   {
      Connection con = null;
      PreparedStatement psGet = null;
      ResultSet rs = null;
      
      try
      {
         con = getDataSource().getConnection();
         psGet = con.prepareStatement(getReadStatement());
         psGet.setObject(1, id);
         rs = psGet.executeQuery();
         if (rs.next())
         {
            Blob blob = rs.getBlob(blobColumn);
            return blob.length();             
         }
         else
         {
            throw new StoreItemNotFoundException("Stream with id: " + id + " does not exist");
         }
      }
      catch (SQLException e)
      {
         throw new StoreException(e);
      }
      finally
      {
         DbUtils.closeQuietly(rs);
         DbUtils.closeQuietly(psGet);
         DbUtils.closeQuietly(con);
      }      
   }
   

   /* (non-Javadoc)
    * @see org.jboss.mail.store.Store#read(java.nio.ByteBuffer)
    */
   public int read(byte[] b, int off, int len, Object id, long position) throws StoreException
   {
      Connection con = null;
      PreparedStatement psGet = null;
      ResultSet rs = null;
      int bytesRead;
      
      try
      {
         con = getDataSource().getConnection();
         psGet = con.prepareStatement(getReadStatement());
         psGet.setObject(1, id);
         rs = psGet.executeQuery();
         if (rs.next())
         {
            Blob blob = rs.getBlob(blobColumn);
            long length = blob.length();
            
            if (length < position)
            {
               return -1;
            }
            
            
            if (useStreams)
            {
               InputStream in = blob.getBinaryStream();
               // Need to skip ahead a few bytes to find the appropriate 
               // offset.
               in.skip(position - 1);
               bytesRead = in.read(b, off, len);
            }
            else
            {
               bytesRead = 0;
               while (len > bytesRead && position <= length)
               {
                  int toPut = Math.min(len - bytesRead, maxBlockSize);
                  byte[] inBytes = blob.getBytes(position, toPut);
                  int numBytes = inBytes.length;
                  System.arraycopy(inBytes, 0, b, off + bytesRead, numBytes);
                  bytesRead += numBytes;
                  position += numBytes;
               }               
            }
         }
         else
         {
            throw new StoreItemNotFoundException("Stream with id: " + id + " does not exist");
         }
      }
      catch (IOException e)
      {
         throw new StoreException(e);
      }
      catch (SQLException e)
      {
         throw new StoreException(e);
      }
      finally
      {
         DbUtils.closeQuietly(rs);
         DbUtils.closeQuietly(psGet);
         DbUtils.closeQuietly(con);
      }
      
      return bytesRead;
   }



   
   /* (non-Javadoc)
    * @see org.jboss.mail.store.Store#read(java.nio.ByteBuffer)
    */
   public ByteBuffer getBuffer(Object id, long position, int len) throws StoreException
   {
      Connection con = null;
      PreparedStatement psGet = null;
      ResultSet rs = null;
      ByteBuffer buffer;
      
      try
      {
         con = getDataSource().getConnection();
         psGet = con.prepareStatement(getReadStatement());
         psGet.setObject(1, id);
         rs = psGet.executeQuery();
         if (rs.next())
         {
            Blob blob = rs.getBlob(blobColumn);
            long length = blob.length();
            
            if (len > maxBlockSize)
            {
               throw new StoreException("len exceeds the maxBlockSize");
            }
            
            if (useStreams)
            {
               InputStream in = blob.getBinaryStream();
               in.skip(position - 1);
               
               int remaining = StoreClerk.longToInt(length - position);
               byte[] b = new byte[Math.min(remaining + 1, len)];
               int numRead = in.read(b);
               buffer = ByteBuffer.wrap(b, 0, numRead);
            }
            else
            {
               byte[] b = blob.getBytes(position, len);
               if (b != null)
               {
                  buffer = ByteBuffer.wrap(b);                                 
               }
               else
               {
                  buffer = ByteBuffer.allocate(0);
               }
            }
         }
         else
         {
            throw new StoreItemNotFoundException("Stream with id: " + id + " does not exist");
         }
      }
      catch (SQLException e)
      {
         throw new StoreException(e);
      }
      catch (IOException e)
      {
         throw new StoreException(e);
      }
      finally
      {
         DbUtils.closeQuietly(rs);
         DbUtils.closeQuietly(psGet);
         DbUtils.closeQuietly(con);
      }
      
      return buffer;
   }   
   
   /**
    * @throws StoreException 
    * @see org.jboss.mail.store.Store#write(java.nio.ByteBuffer)
    */
   public int write(byte[] b, int off, int len, Object id, long position) throws StoreException
   {
      Connection con = null;
      PreparedStatement psGet = null;
      ResultSet rs = null;
      int bytesWritten;
      
      try
      {
         con = getDataSource().getConnection();
         psGet = con.prepareStatement(getReadStatement());
         psGet.setObject(1, id);
         rs = psGet.executeQuery();
         if (rs.next())
         {
            Blob blob = rs.getBlob(blobColumn);
            
            if (useStreams)
            {
               OutputStream out = blob.setBinaryStream(position);
               out.write(b, off, len);
               bytesWritten = len;
            }
            else
            {
               bytesWritten = 0;
               while (len > bytesWritten)
               {
                  int toGet = Math.min(len - bytesWritten, maxBlockSize);
                  blob.setBytes(position, b, off + bytesWritten, toGet);
                  bytesWritten += toGet;
                  position += toGet;
               }               
            }
         }
         else
         {
            throw new StoreItemNotFoundException("Stream with id: " + id + " does not exist");
         }
      }
      catch (SQLException e)
      {
         throw new StoreException(e);
      }
      catch (IOException e)
      {
         throw new StoreException(e);
      }
      finally
      {
         DbUtils.closeQuietly(rs);
         DbUtils.closeQuietly(psGet);
         DbUtils.closeQuietly(con);
      }      
      
      return bytesWritten;
   }
   
   
   /**
    * Loads the metadata for object defined by id.
    * 
    * @see org.jboss.mail.store.Store#read(java.nio.ByteBuffer)
    */
   public StoreItemMetaData getStoreItemMetaData(Object id) throws StoreException
   {
      Connection con = null;
      PreparedStatement psGet = null;
      ResultSet rs = null;
      StoreItemMetaData meta;
      
      try
      {
         con = getDataSource().getConnection();
         psGet = con.prepareStatement(getReadStatement());
         psGet.setObject(1, id);
         rs = psGet.executeQuery();
         if (rs.next())
         {
            Blob blob = rs.getBlob(blobColumn);
            long length = blob.length();
            boolean compressed = TRUE.equals(rs.getString(compressedColumn));
            meta = new StoreItemMetaData(id, length, compressed, 1L);
         }
         else
         {
            throw new StoreItemNotFoundException("Stream with id: " + id + " does not exist");
         }
      }
      catch (SQLException e)
      {
         throw new StoreException(e);
      }
      finally
      {
         DbUtils.closeQuietly(rs);
         DbUtils.closeQuietly(psGet);
         DbUtils.closeQuietly(con);
      }
      
      return meta;
   }   
   
   

   
   /**
    * Blobs are index from 1.
    * 
    * @see org.jboss.mail.store.Store#getStartIndex()
    */
   public long getStartIndex()
   {
      // TODO Auto-generated method stub
      return 1L;
   }
   
   /**
    * Generic command processing method.
    * 
    * @param id
    * @param cmd
    * @throws StoreException
    */
   public void processSelect(Object id, BlobCommand cmd) throws StoreException
   {
      Connection cn = null;
      PreparedStatement ps = null;
      ResultSet rs = null;
      
      try
      {
         cn = getDataSource().getConnection();
         ps = cn.prepareStatement(getReadStatement());
         ps.setObject(1, id);
         rs = ps.executeQuery();
         if (rs.next())
         {
            cmd.process(rs);
         }
         else
         {
            throw new StoreItemNotFoundException("Stream with id: " + id + " does not exist");
         }         
      }
      catch (SQLException e)
      {
         throw new StoreException(e);
      }
      finally
      {
         if (cmd.closeConnection())
         {
            DbUtils.closeQuietly(rs);
            DbUtils.closeQuietly(ps);
            DbUtils.closeQuietly(cn);            
         }
      }
   }
   


   /**
    * Creates the appropriate input
    * 
    * @see org.jboss.mail.store.Store#getInputStream(java.lang.Object)
    */
   public InputStream getInputStream(Object id) throws StoreException
   {
      InputStream in;
      InputStream blobIn;

      StoreItemMetaData meta = getStoreItemMetaData(id);
      if (disconnected)
      {
         blobIn = new BlobInputStream(this, meta, getMinBufferSize(), getMaxBufferSize());         
      }
      else
      {
         GetInputStreamCommand cmd = new GetInputStreamCommand();
         processSelect(id, cmd);
         blobIn = cmd.getInputStream();
      }
      if (meta.isCompressed())
      {
         try
         {
            in = new GZIPInputStream(blobIn, getCompressBufferSize());
         }
         catch (IOException e)
         {
            throw new StoreException("Unable to create compressed stream for id =  " + id, e);
         }
      }
      else
      {
         in = blobIn;
      }
      
      return in;
   }


   /**
    * @throws StoreException 
    * @see org.jboss.mail.store.Store#getOuputStream(java.lang.Object)
    */
   public OutputStream getOuputStream(Object id) throws StoreException
   {
      OutputStream out;
      StoreItemMetaData meta = getStoreItemMetaData(id);
      OutputStream blobOut;
      
      if (isDisconnected())
      {
         blobOut = new BlobOutputStream(this, meta, getBufferSize());         
      }
      else
      {
         GetOutputStreamCommand cmd = new GetOutputStreamCommand();
         processSelect(id, cmd);
         blobOut = cmd.getOutputStream();
      }
      if (meta.isCompressed())
      {
         try
         {
            out = new GZIPOutputStream(blobOut, getCompressBufferSize());
         }
         catch (IOException e)
         {
            throw new StoreException("Unable to create compressed stream for id = " + id, e);
         }
      }
      else
      {
         out = blobOut;
      }
      
      return out;
   }



   /* (non-Javadoc)
    * @see org.jboss.mail.store.Store#createStoreItem(java.lang.Object)
    */
   public StoreItem createStoreItem() throws StoreException
   {
      Object id = createId();
      Connection con = null;
      PreparedStatement pstmt = null;
      StoreItem item;
      
      try
      {
         con = getDataSource().getConnection();
         pstmt = con.prepareStatement(getInsertStatement());
         pstmt.setObject(1, id);
         pstmt.setObject(2, (isCompress() ? "T" : "F"));
         int i = pstmt.executeUpdate();
         if (i > 0)
         {
            item = new StoreItemImpl(id, this);
         }
         else
         {
            throw new StoreException("Unable to create stream store for id = " + id);
         }
      }
      catch (SQLException e)
      {
         throw new StoreException("Unable to create stream store for id = " + id, e);
      }
      finally
      {
         DbUtils.closeQuietly(pstmt);
         DbUtils.closeQuietly(con);
      }
      
      return item;
   }

   
   public void validate() throws StoreException
   {
      super.validate();
      testNull("createStatement");
      testNull("readStatement");
      testNull("insertStatement");
      testNull("writeStatement");
      testNull("idColumn");
      testNull("blobColumn");
      testNull("compressedColumn");
      testNull("dataSource");
   }
   
   

   /**
    * Creates the databse table if necessary.
    * 
    * @see org.jboss.mail.store.AbstractStore#init()
    */
   public void init() throws StoreException
   {
      Connection cn = null;
      ResultSet rsTables = null;
      ResultSet rsColumns = null;
      PreparedStatement ps = null;
      
      try
      {
         cn = getDataSource().getConnection();
         DatabaseMetaData dbmd = cn.getMetaData();
         rsTables = dbmd.getTables(null, null, getTableName(), null);
         
         if (rsTables.next())
         {
            log.info("Store table found");
            // TODO: Check all of the column in table have the
            // correct names and types.
         }
         else
         {
            log.info("Database table for store is not found creating: " + getCreateStatement());
            ps = cn.prepareStatement(getCreateStatement());
            ps.execute();
            // Some databases don't return true on a create.
            // Assume if no exception thrown create was OK.
            // TODO: Check creation afterwards.
         }
         
      }
      catch (SQLException e)
      {
         throw new StoreException("Unable to read database metadata", e);
      }
      finally
      {
         DbUtils.closeQuietly(rsTables);
         DbUtils.closeQuietly(cn);
         DbUtils.closeQuietly(cn);
      }
   }

   

   /**
    * Command for getting the InputStream from a blob.
    *
    */
   class GetInputStreamCommand extends BlobCommand
   {
      InputStream in;

      /* (non-Javadoc)
       * @see org.jboss.mail.store.jdbc3.BlobCommand#process(java.sql.ResultSet)
       */
      public void process(ResultSet rs) throws StoreException
      {
         try
         {
            Connection cn = rs.getStatement().getConnection();
            Blob b = rs.getBlob(getBlobColumn());
            in = new WrappedInputStream(b.getBinaryStream(), cn);
         }
         catch (SQLException e)
         {
            throw new StoreException(e);
         }
      }
      
      public InputStream getInputStream()
      {
         return in;
      }
      
      /**
       * We want the connection to remain open.  It will closed
       * when the stream's <codde>close()</code> method is called
       */
      public boolean closeConnection()
      {
         return false;
      }
      
      public String getStatement()
      {
         return getReadStatement();
      }
   }
   
   /**
    * Command for getting the output stream from a blob.
    *
    */
   class GetOutputStreamCommand extends BlobCommand
   {
      OutputStream out;

      /* (non-Javadoc)
       * @see org.jboss.mail.store.jdbc3.BlobCommand#process(java.sql.ResultSet)
       */
      public void process(ResultSet rs) throws StoreException
      {
         try
         {
            Connection cn = rs.getStatement().getConnection();
            Blob b = rs.getBlob(getBlobColumn());
            out = new WrappedOutputStream(b.setBinaryStream(1L), cn);
         }
         catch (SQLException e)
         {
            throw new StoreException(e);
         }
      }
      
      public OutputStream getOutputStream()
      {
         return out;
      }
      
      /**
       * We want the connection to remain open.  It will closed
       * when the stream's <codde>close()</code> method is called
       */      
      public boolean closeConnection()
      {
         return false;
      }
      
      public String getStatement()
      {
         return getWriteStatement();
      }
   }
   
}
