/*
 * Created on Jan 23, 2005
 *
 */
package org.jboss.mail.store.postgresql;

import java.io.InputStream;
import java.io.OutputStream;
import java.nio.ByteBuffer;
import java.sql.Connection;
import java.sql.SQLException;

import javax.sql.DataSource;

import org.apache.commons.dbutils.DbUtils;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.jboss.mail.store.AbstractStore;
import org.jboss.mail.store.BlobInputStream;
import org.jboss.mail.store.BlobOutputStream;
import org.jboss.mail.store.StoreClerk;
import org.jboss.mail.store.StoreException;
import org.jboss.mail.store.StoreItem;
import org.jboss.mail.store.StoreItemImpl;
import org.jboss.mail.store.StoreItemMetaData;
import org.postgresql.PGConnection;
import org.postgresql.largeobject.LargeObject;
import org.postgresql.largeobject.LargeObjectManager;

/**
 * @author Michael Barker <mailto:mike@middlesoft.co.uk>
 *
 */
public class PosgreSQLStore extends AbstractStore
{
   private DataSource dataSource;
   private static Log log = LogFactory.getLog(PosgreSQLStore.class);
   private boolean disconnected = false;
   

   /**
    * @return Returns the dataSource.
    */
   public DataSource getDataSource()
   {
      return dataSource;
   }
   /**
    * @param dataSource The dataSource to set.
    */
   public void setDataSource(DataSource dataSource)
   {
      this.dataSource = dataSource;
   }
   
   
   /**
    * @return Returns the disconnected.
    */
   public boolean isDisconnected()
   {
      return disconnected;
   }
   /**
    * @param disconnected The disconnected to set.
    */
   public void setDisconnected(boolean disconnected)
   {
      this.disconnected = disconnected;
   }
   
   
   /* (non-Javadoc)
    * @see org.jboss.mail.store.AbstractStore#init()
    */
   public void init() throws StoreException
   {
      // Nothing to initalise.
      // TODO: Perhaps check database connection.
   }

   /**
    * Reads some bytes from the store object
    * 
    * @see org.jboss.mail.store.Store#read(byte[], int, int, java.lang.Object, long)
    */
   public int read(byte[] b, int off, int len, Object id, long position) throws StoreException
   {
      Connection cn = null;
      LargeObject blob = null;
      int numRead;
      int oid = ((Integer) id).intValue();
      
      try
      {
         cn = getDataSource().getConnection();
         cn.setAutoCommit(false);
         LargeObjectManager mgr = ((PGConnection) cn).getLargeObjectAPI();
         blob = mgr.open(oid);
         blob.seek(StoreClerk.longToInt(position));
         numRead = blob.read(b, off, len);
      }
      catch (SQLException e)
      {
         throw new StoreException(e);
      }
      finally
      {
         PGUtil.closeQuietly(blob);
         DbUtils.closeQuietly(cn);
      }
      
      return numRead;
   }

   /**
    * @see org.jboss.mail.store.Store#write(byte[], int, int, java.lang.Object, long)
    */
   public int write(byte[] b, int off, int len, Object id, long position) throws StoreException
   {
      Connection cn = null;
      LargeObject blob = null;
      int oid = ((Integer) id).intValue();
      
      try
      {
         cn = getDataSource().getConnection();
         cn.setAutoCommit(false);
         LargeObjectManager mgr = ((PGConnection) cn).getLargeObjectAPI();
         blob = mgr.open(oid);
         blob.seek(StoreClerk.longToInt(position));
         blob.write(b, off, len);
         cn.commit();
      }
      catch (SQLException e)
      {
         throw new StoreException(e);
      }
      finally
      {
         DbUtils.closeQuietly(cn);
      }
      
      return len;
   }


   /* (non-Javadoc)
    * @see org.jboss.mail.store.Store#getBuffer(java.lang.Object, long, int)
    */
   public ByteBuffer getBuffer(Object id, long position, int len) throws StoreException
   {
      Connection cn = null;
      LargeObject blob = null;
      int oid = ((Integer) id).intValue();
      ByteBuffer buffer;
      
      try
      {
         cn = getDataSource().getConnection();
         cn.setAutoCommit(false);
         LargeObjectManager mgr = ((PGConnection) cn).getLargeObjectAPI();
         blob = mgr.open(oid);
         blob.seek(StoreClerk.longToInt(position));
         byte[] b = new byte[len];
         int numRead = blob.read(b, 0, len);
         buffer = ByteBuffer.wrap(b, 0, numRead);
      }
      catch (SQLException e)
      {
         throw new StoreException(e);
      }
      finally
      {
         PGUtil.closeQuietly(blob);
         DbUtils.closeQuietly(cn);
      }
      
      return buffer;
   }

   /**
    * Creates a postgres large object for use as a store item.
    * 
    * @see org.jboss.mail.store.Store#createStoreItem(java.lang.Object)
    */
   public StoreItem createStoreItem() throws StoreException
   {
      Connection cn = null;
      StoreItem item;
      
      try
      {
         cn = getDataSource().getConnection();
         cn.setAutoCommit(false);
         LargeObjectManager mgr = ((PGConnection) cn).getLargeObjectAPI();
         int oid = mgr.create();
         //mgr.open(oid);
         item = new StoreItemImpl(new Integer(oid), this);
         cn.commit();
      }
      catch (SQLException e)
      {
         throw new StoreException(e);
      }
      finally
      {
         DbUtils.closeQuietly(cn);
      }
      
      return item;
   }
   
   /**
    * Returns an input stream to the blob, that will also close the
    * associated connection when the stream is closed.
    * 
    * @param id
    * @return
    * @throws StoreException
    */
   private InputStream getBinaryStream(Object id) throws StoreException
   {
      Connection cn = null;
      int oid = ((Integer) id).intValue();
      InputStream in;
      
      try
      {
         cn = getDataSource().getConnection();
         cn.setAutoCommit(false);
         LargeObjectManager mgr = ((PGConnection)cn).getLargeObjectAPI();
         LargeObject blob = mgr.open(oid);
         in = new PGWrappedInputStream(blob.getInputStream(), cn, blob);
      }
      catch (SQLException e)
      {
         throw new StoreException(e);
      }
      
      return in;
   }
   
   /**
    * Returns and output stream to the blob that will also close the 
    * associated connection when the stram is closed.
    * 
    * @param id
    * @return
    * @throws StoreException
    */
   private OutputStream setBinaryStream(Object id) throws StoreException
   {
      Connection cn = null;
      int oid = ((Integer) id).intValue();
      OutputStream out;
      
      try
      {
         cn = getDataSource().getConnection();
         cn.setAutoCommit(false);
         LargeObjectManager mgr = ((PGConnection)cn).getLargeObjectAPI();
         LargeObject blob = mgr.open(oid);
         out = new PGWrappedOutputStream(blob.getOutputStream(), cn, blob);
      }
      catch (SQLException e)
      {
         throw new StoreException(e);
      }
      
      return out;
   }
   
   
   /**
    * TODO: Add compression.
    * 
    * @see org.jboss.mail.store.Store#getInputStream(java.lang.Object)
    */
   public InputStream getInputStream(Object id) throws StoreException
   {
      if (isDisconnected())
      {
         return new BlobInputStream(this, getStoreItemMetaData(id), getMaxBufferSize(), getMinBufferSize());
      }
      else
      {
         return getBinaryStream(id);
      }
   }

   /**
    * TODO: Add compression.
    * 
    * @see org.jboss.mail.store.Store#getInputStream(java.lang.Object)
    */
   public OutputStream getOuputStream(Object id) throws StoreException
   {
      if (isDisconnected())
      {
         return new BlobOutputStream(this, getStoreItemMetaData(id), getBufferSize());
      }
      else
      {
         return setBinaryStream(id);
      }
   }

   /**
    * Gets the metadata for this object.
    * 
    * @see org.jboss.mail.store.Store#getStoreItemMetaData(java.lang.Object)
    */
   public StoreItemMetaData getStoreItemMetaData(Object id) throws StoreException
   {
      Connection cn = null;
      LargeObject blob = null;
      int oid = ((Integer) id).intValue();
      StoreItemMetaData meta;
      
      try
      {
         cn = getDataSource().getConnection();
         cn.setAutoCommit(false);
         LargeObjectManager mgr = ((PGConnection)cn).getLargeObjectAPI();
         blob = mgr.open(oid);
         int size = blob.size();
         // TODO: Does this LO start at 0.
         meta = new StoreItemMetaData(id, size, false, 0);
      }
      catch (SQLException e)
      {
         throw new StoreException(e);
      }
      finally
      {
         PGUtil.closeQuietly(blob);
         DbUtils.closeQuietly(cn);
      }
      
      return meta;
   }
   
   public void validate() throws StoreException
   {
      super.validate();
      testNull("dataSource");
   }   

}
