/*
 * Created on Jan 3, 2005
 *
 */
package org.jboss.mail.store;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.sql.SQLException;
import java.util.Arrays;

import javax.naming.NamingException;

import junit.framework.TestCase;

/**
 * @author Michael Barker <mailto:mike@middlesoft.co.uk>
 *
 */
public abstract class TestStore extends TestCase
{
   
   public abstract Store getStore() throws StoreException;
   
   public void testStoreByteArray01() throws StoreException
   {
      int bufSize = 1024;

      Store store = getStore();
      
      StoreItem item = store.createStoreItem();
      Object id = item.getId();
      StoreItemMetaData meta = store.getStoreItemMetaData(id);

      byte[] outBuf = new byte[bufSize];

      for (int i = 0; i < outBuf.length; i++)
      {
         byte b = (byte) ((i % 26) + 97);
         outBuf[i] = b;
      }
      
      long t0 = System.currentTimeMillis();
      store.write(outBuf, 0, outBuf.length, id, meta.getStartIndex());
      long t1 = System.currentTimeMillis();
      
      System.out.println("Time to write: " + (t1 - t0));
      
      byte[] inBuf = new byte[bufSize];
      long t2 = System.currentTimeMillis();
      store.read(inBuf, 0, inBuf.length, id, meta.getStartIndex());
      long t3 = System.currentTimeMillis();
      
      System.out.println("Time to read: " + (t3 - t2));

      assertTrue(Arrays.equals(inBuf, outBuf));
   }
   

   public void testStoreByteArray02() throws StoreException
   {

      Store store = getStore();
      
      StoreItem item = store.createStoreItem();
      Object id = item.getId();
      StoreItemMetaData meta = store.getStoreItemMetaData(id);

      byte[] outBuf = new byte[256];
      
      for (int i = 0; i < outBuf.length; i++)
      {
         outBuf[i] = (byte) (i & 0x000000ff);
      }
      
      store.write(outBuf, 0, outBuf.length, id, meta.getStartIndex());
      
      byte[] inBuf = new byte[outBuf.length];
      store.read(inBuf, 0, inBuf.length, id, meta.getStartIndex());

      
      for (int i = 0; i < outBuf.length; i++)
      {
         if (outBuf[i] != inBuf[i])
         {
            System.out.println("in: " + outBuf[i] + ", out: " + inBuf[i]);            
         }
      }
      
      
      assertTrue(Arrays.equals(inBuf, outBuf));
   }
   
   /**
    * Write a single byte to the output stream and read it back again.
    * @throws StoreException 
    * @throws SQLException 
    * @throws NamingException 
    * 
    * @throws StoreItemNotFoundException
    * @throws NamingException
    * @throws SQLException
    * @throws IOException 
    * @throws IOException
    * @throws Exception 
    */
   public void testBlobWriteByte() throws Exception
   {
      int blockSize = 8 * 1024;
      
      StoreItem item = getStoreItem();
      OutputStream out = item.getOuputStream();      
      
      int value = 'a';
      out.write('a');
      out.flush();
      out.close();
      
      InputStream in = item.getInputStream();
      int result = in.read();
      
      assertEquals(value, result);
   }

   /**
    * @return
    * @throws StoreException 
    */
   private StoreItem getStoreItem() throws StoreException
   {
      return getStore().createStoreItem();
   }

   /**
    * Write an array of bytes to the output stream and read them back again.
    * @throws StoreItemNotFoundException
    * @throws NamingException
    * @throws SQLException
    * @throws IOException
    * @throws Exception 
    */
   public void testBlobWriteByteArrayTwiceBufferSize() throws Exception
   {
      int streamBufferSize = 1024;
      int storeBufferSize = 1024 * 1024;
      
      StoreItem item = getStoreItem();
      InputStream in = item.getInputStream();
      OutputStream out = item.getOuputStream();      
      
      byte[] bs = new byte[2048];
      
      for (int i = 0; i < bs.length; i++)
      {
         bs[i] = (byte) ((i % 26) + 97);
      }
      
      out.write(bs);
      out.flush();
      out.close();
      
      byte retBs[] = new byte[2048];
      int numRead = in.read(retBs);

      for(int i = 0; i < bs.length && i < retBs.length; i++)
      {
         assertEquals("Index " + i + " is not the same.", bs[i], retBs[i]);
      }
      
      assertEquals(2048, numRead);
      assertTrue(Arrays.equals(bs, retBs));
   }
   
   
   
   
   /**
    * Write an array of bytes to the output stream and read them back again.
    * 
    * The size of the input will not be a multiple of the size of the buffer.
    * @throws Exception 
    * 
    * @throws NamingException
    * @throws SQLException
    */
   public void testBlobWriteByteArrayNotMultipleOfBufferSize() throws Exception
   {
      int streamBufferSize = 1024;
      int storeBufferSize = 1024 * 1024;
      
      StoreItem item = getStoreItem();
      InputStream in = item.getInputStream();
      OutputStream out = item.getOuputStream();      
      
      byte[] bs = new byte[2048];
      
      for (int i = 0; i < bs.length; i++)
      {
         bs[i] = (byte) ((i % 26) + 97);
      }
      
      out.write(bs);
      out.flush();
      out.close();
      
      byte retBs[] = new byte[2048];
      int numRead = in.read(retBs);

      for(int i = 0; i < bs.length && i < retBs.length; i++)
      {
         if (bs[i] != retBs[i])
         {
            assertFalse("Index " + i + " is not the same.", true);
            break;
         }
      }
      
      assertEquals(2048, numRead);
      assertTrue(Arrays.equals(bs, retBs));
   }   
   
   
   /**
    * Write an array of bytes to the output stream and read them back again.
    * @throws Exception 
    * 
    * @throws NamingException
    */
   public void testBlob10MB() throws Exception
   {
      int bufSize = 10 * 1024 * 1024; // 10 MB

      StoreItem item = getStoreItem();
      InputStream in = item.getInputStream();
      OutputStream out = item.getOuputStream();      
      
      
      byte[] bs = new byte[bufSize];
      
      for (int i = 0; i < bs.length; i++)
      {
         bs[i] = (byte) ((i % 26) + 97);
      }
      
      
      long t0 = System.currentTimeMillis();
      out.write(bs);
      out.flush();
      out.close();
      long t1 = System.currentTimeMillis();
      
      byte retBs[] = new byte[bufSize];
      long t2 = System.currentTimeMillis();
      int numRead = in.read(retBs);
      long t3 = System.currentTimeMillis();

      for(int i = 0; i < bs.length && i < retBs.length; i++)
      {
         if (bs[i] != retBs[i])
         {
            assertFalse("Index " + i + " is not the same.", true);
            break;
         }
      }
      
      assertEquals(bufSize, numRead);
      assertTrue(Arrays.equals(bs, retBs));
      
      System.out.println("Time to write: " + (t1 - t0));
      System.out.println("Time to read: " + (t3 - t2));
   }   
   
   /**
    * Write an array of bytes to the output stream and read them back again.
    * @throws Exception 
    * 
    * @throws NamingException
    */
   public void testBlob1MB() throws Exception
   {
      int bufSize = 1 * 1024 * 1024; // 10 MB

      StoreItem item = getStoreItem();
      InputStream in = item.getInputStream();
      OutputStream out = item.getOuputStream();      
      
      
      byte[] bs = new byte[bufSize];
      
      for (int i = 0; i < bs.length; i++)
      {
         bs[i] = (byte) ((i % 26) + 97);
      }
      
      
      long t0 = System.currentTimeMillis();
      for (int i = 0; i < 10; i++)
      {
         out.write(bs);
      }
      out.flush();
      out.close();
      long t1 = System.currentTimeMillis();
      
      byte retBs[] = new byte[bufSize];
      long t2 = System.currentTimeMillis();
      for (int i = 0; i < 10; i++)
      {
         int numRead = in.read(retBs);         
         assertFalse(numRead == -1 && i < 9);
      }
      long t3 = System.currentTimeMillis();

      System.out.println("Time to write: " + (t1 - t0));
      System.out.println("Time to read: " + (t3 - t2));
   }      
   
   /**
    * Write an array of bytes to the output stream and read them back again.
    * @throws Exception 
    * 
    * @throws NamingException
    */
   public void testBlob128MB() throws Exception
   {
      int bufSize = 1 * 1024 * 1024; // 1 MB
      int loopSize = 128; // Write buffer this many times.

      StoreItem item = getStoreItem();
      InputStream in = item.getInputStream();
      OutputStream out = item.getOuputStream();      
      
      
      byte[] bs = new byte[bufSize];
      
      for (int i = 0; i < bs.length; i++)
      {
         bs[i] = (byte) ((i % 26) + 97);
      }
      
      
      long t0 = System.currentTimeMillis();
      for (int i = 0; i < loopSize; i++)
      {
         out.write(bs);
      }
      out.flush();
      out.close();
      long t1 = System.currentTimeMillis();
      
      byte retBs[] = new byte[bufSize];
      int numRead = 0;
      long t2 = System.currentTimeMillis();
      for (int i = 0; i < loopSize; i++)
      {
         numRead += in.read(retBs);
         //assertFalse(numRead == -1 && i < (loopSize - 1));
      }
      long t3 = System.currentTimeMillis();

      assertEquals(loopSize * bufSize, numRead);
      
      System.out.println("Time to write: " + (t1 - t0));
      System.out.println("Time to read: " + (t3 - t2));
   }       
   
   
   
}
